!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! TELCAL type definition modules, initialization and copy routines
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
module fit_definitions
  ! EXPORT
  !----------------------------------------------------------------------
  ! TELCAL support module 
  ! Purpose: Definitions of FIT derived types
  !----------------------------------------------------------------------  
  !
  type :: simple_1d ! Data vector (simple precision, 1d)
     integer :: n                 ! Number of data
     real(8), pointer :: x(:)     ! Abcissae (in)
     real(8), pointer :: y(:)     ! Values   (in)
     real(8), pointer :: w(:)     ! Weights  (in)
     real(8), pointer :: d(:)     ! derivative of x with time, Not always allocated
  end type simple_1d
  !
  type :: fit_par ! Description of a fitted parameter
     character(len=32) :: name ! Parameter name          (in)
     real(8) :: guess          ! Guess                   (in)
     real(8) :: value          ! Fit value               (out)
     real(8) :: error          ! Fit uncertainty         (out)
     real(8) :: mini           ! Minimum possible value  (in)
     real(8) :: maxi           ! Maximum possible value  (in)
     logical :: fixed          ! Is the parameter fixed? (in)
     logical :: padding        ! Padding for structure alignment
  end type fit_par
  !
  type :: fit_fun ! Description of a fitted function
     character(len=16) :: name         ! Fitted function name (eg GAUSSIAN)
     character(len=16) :: method       ! Fitting library (eg SLATEC)
     real    :: chi2                   ! Fit chi square
     real    :: rms                    ! Residual RMS
     integer :: flag                   ! Error/Quality code
     integer :: ncall                  ! Number of iterations
     integer :: npar                   ! Number of fitted parameters
     type (fit_par), pointer :: par(:) ! Array of parameter descriptions
  end type fit_fun
  !
  type :: fit_var ! Variable fitted parameters only
     integer :: n               ! Number
     real(8), pointer :: x(:)   ! Values
     real(8), pointer :: d(:)   ! Errors 
     integer, pointer :: idx(:) ! Position index inside full parameter vector
  end type fit_var
  !
end module fit_definitions
!
subroutine null_simple_1d(dat)
  ! EXPORT
  use fit_definitions
  !----------------------------------------------------------------------
  ! TELCAL support routine of its derived types
  ! Purpose: Initialize a simple precision data vector
  !----------------------------------------------------------------------
  type (simple_1d), intent(inout) :: dat ! Data vector
  !
  ! Local variables
  dat%n = 0
  if (associated(dat%x)) deallocate(dat%x)
  if (associated(dat%y)) deallocate(dat%y)
  if (associated(dat%w)) deallocate(dat%w)
  if (associated(dat%d)) deallocate(dat%d)
  nullify(dat%x)
  nullify(dat%y)
  nullify(dat%w)
  nullify(dat%d)
end subroutine null_simple_1d
!
subroutine null_parameter(par)
  ! EXPORT
  use fit_definitions
  !----------------------------------------------------------------------
  ! TELCAL support routine of its derived types
  ! Purpose: Initialize a fitted parameter
  !----------------------------------------------------------------------
  type (fit_par), intent(inout) :: par ! Fitted parameter
  !
  ! Local variables
  par%guess = 0
  par%value = 0
  par%error = 0
  par%mini = -1e30
  par%maxi =  1e30
  par%fixed = .false.
  par%name = 'UNDEFINED'
end subroutine null_parameter
!
subroutine null_function(fun)
  ! EXPORT
  use fit_definitions
  !----------------------------------------------------------------------
  ! TELCAL support routine of its derived types
  ! Purpose: Initialize a fitted function description
  !----------------------------------------------------------------------
  type (fit_fun), intent(inout) :: fun ! Description of the fitted function
  !
  ! Local variables
  fun%name   = 'UNDEFINED'
  fun%method = 'UNDEFINED'
  fun%chi2   = 0.
  fun%rms    = 0.
  fun%flag   = 1 ! Waiting for data
  fun%ncall  = 0
  fun%npar   = 0
  if (associated(fun%par)) deallocate(fun%par)
  nullify(fun%par)
end subroutine null_function
!
! Fit derived type copy routines
!
subroutine copy_simple_1d(in,out)
  use fit_definitions
  !----------------------------------------------------------------------
  ! TELCAL support routine of its derived types
  ! Purpose: Copy a simple precision data vector
  !----------------------------------------------------------------------
  type (simple_1d), intent(in)  :: in  ! Original vector
  type (simple_1d), intent(out) :: out ! Copied vector
  !
  ! Local variables
  out%n = in%n
  if (associated(in%x).and.associated(out%x)) out%x = in%x  
  if (associated(in%y).and.associated(out%y)) out%y = in%y
  if (associated(in%w).and.associated(out%w)) out%w = in%w
  if (associated(in%d).and.associated(out%d)) out%d = in%d
  !
end subroutine copy_simple_1d
!
subroutine copy_function(in,out)
  use fit_definitions
  !----------------------------------------------------------------------
  ! TELCAL support routine of its derived types
  ! Purpose: Copy the description of a fitted function
  !----------------------------------------------------------------------
  type (fit_fun), intent(in)  :: in  ! Original function
  type (fit_fun), intent(out) :: out ! Copied function
  !
  ! Local variables
  integer i
  !
  out%name   = in%name  
  out%method = in%method
  out%chi2   = in%chi2
  out%rms    = in%rms   
  out%flag   = in%flag  
  out%ncall  = in%ncall 
  out%npar   = in%npar
  do i=1,in%npar
     call null_parameter(out%par(i))
     out%par(i) = in%par(i)
  enddo
  !
end subroutine copy_function
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! User interface derived type definitions
!
module user_guess_definitions
  !----------------------------------------------------------------------
  ! TELCAL support module 
  ! Purpose: Definition of types linked to user guess of fit parameters
  !----------------------------------------------------------------------  
  type :: guess_t
     real(8) :: value ! Fit value
     logical :: fixed ! Is the parameter fixed?
     logical :: padding
  end type guess_t
end module user_guess_definitions
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! POINTING derived type definitions
!
module point_definitions
  use fit_definitions
  use user_guess_definitions
  !----------------------------------------------------------------------
  ! TELCAL support module for pointing
  !----------------------------------------------------------------------
  !
  type :: point_guess_t
     type (guess_t) :: offset       ! 1st order baseline offset
     type (guess_t) :: slope        ! 1st order baseline slope
     type (guess_t) :: area         ! 1st Gaussian area
     type (guess_t) :: position     ! 1st Gaussian position
     type (guess_t) :: width        ! 1st Gaussian width
     type (guess_t) :: beam_throw   ! Beam  throw in dual beam case
     type (guess_t) :: area_ratio   ! Area  ratio in dual beam case
     type (guess_t) :: width_ratio  ! Width ratio in dual beam case
  end type point_guess_t
  !
  type :: pointing_t
     integer :: sys           ! Scanning coordinate system
     character(16)    :: dir  ! Scanning directions
     character(16)    :: type ! Pointing kind (SINGLE or DUAL)
     type (simple_1d) :: dat  ! Data vectors
     type (simple_1d) :: sol  ! Solution vectors
     type (fit_fun)   :: fun  ! Fit functions
  end type pointing_t
  !
end module point_definitions
!
module pcross_definitions
  use point_definitions
  !----------------------------------------------------------------------
  ! TELCAL support module for pointing through a cross
  ! The pcross array is saved because a SIC structure will be mapped on it
  !----------------------------------------------------------------------
  !
  integer, parameter :: mcross = 16         ! Maximum number of cross subscans
  integer            :: ncross              ! Actual  number of cross subscans
  type (pointing_t), save :: pcross(mcross) ! Array of pointing structures
  !  
end module pcross_definitions
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! FOCUS derived type definitions
!
module focus_definitions
  use fit_definitions
  !----------------------------------------------------------------------
  ! TELCAL support module for focus
  !----------------------------------------------------------------------
  !
  type :: focus
     integer :: sys             ! Scanning coordinate system
     character(16)    :: dir(2) ! Scanning directions
     type (simple_1d) :: dat(2) ! Data vectors
     type (simple_1d) :: sol(2) ! Solution vectors
     type (fit_fun)   :: fun(2) ! Fit functions
  end type
  !
  type (focus), save :: zfocus
  !
end module focus_definitions
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! CHOPPER derived type definitions
!
module chopper_definitions
  !----------------------------------------------------------------------
  ! TELCAL support module for chopper wheel calibration
  !----------------------------------------------------------------------
  !
  type :: chop_mode ! 
     logical :: water ! Search for water?
     logical :: atm   ! Compute opacities?
     logical :: trec  ! Compute trec?
     logical :: tcal  ! Compute tcal?
     logical :: tsys  ! Compute tsys?
  end type chop_mode
  !
  type :: telescope !
     real (8) :: alti ! [  km] Telescope altitude
     real (8) :: lati ! [ rad] Telescope latitude    
     real (8) :: elev ! [ rad] Telescope elevation
     real (8) :: hwat ! [  km] Water scale height
     real (8) :: tout ! [   K] Outside temperature
     real (8) :: tcab ! [   K] Cabin   temperature 
     real (8) :: pres ! [mbar] Ambiant pressure
  end type telescope
  !
  type :: dsb_value ! (e.g. frequency, temperature)
     real (8) :: s ! Signal value   
     real (8) :: i ! Image  value
  end type dsb_value
  !
  type :: mixer ! Comment: feff+fout+fcab = 1.0 
     real (8) :: sbgr  ! [   ] Sideband gain ratio
     real (8) :: beff  ! [   ] Beam efficiency    
     real (8) :: feff  ! [   ] Forward efficiency 
     real (8) :: fout  ! [   ] Coupling to ground 
     real (8) :: fcab  ! [   ] Coupling to cabin  
     real (8) :: temp  ! [  K] Noise temperature
  end type mixer
  !
  type :: chop_load ! Load description
     real (8) :: count ! [ ] Power expressed in counts
     real (8) :: temp  ! [K] Power expressed in Rayleigh-Jeans temperature
     real (8) :: ceff  ! [ ] Coupling. Comment: 1-coupling = sky coupling
  end type chop_load
  !
  type :: chop_meas ! Measurements to perform a "chopper" calibration
     real (8)         :: dark_count ! [ ] Dark power expressed in counts
     real (8)         ::  sky_count ! [ ] Sky  power expressed in counts
     type (chop_load) :: cold       ! [ ] Cold load description
     type (chop_load) :: hot        ! [ ] Hot  load description
  end type chop_meas
  !
  integer, parameter :: msky=10
  type :: skydip_meas ! Measurements to perform a skydip reduction
     type (dsb_value) :: freq         ! [MHz] Signal and Image frequencies
     integer          :: nsky         ! [   ] Number of sky measurements
     real (8)         :: elev(msky)   ! [rad] Elevations
     type (chop_load) :: emiss(msky)  !       Power (counts and temp) measured
     integer          :: nhot         ! [   ] Number of hot load measurements
     type (chop_load) :: hot(msky)    !       Power (counts and temp) on hot load (chopper)
     integer          :: ncold        ! [   ] Number of cold load measurements
     type (chop_load) :: cold(msky)   !       Power (counts and temp) on cold load
  end type skydip_meas
  !
  type :: opacity !
     real (8) :: wat ! [neper] Opacity due to water
     real (8) :: oth ! [neper] Opacity due to other components
     real (8) :: tot ! [neper] Total opacity
  end type opacity
  !
  type :: atm_prop ! Atmospheric properties
     real (8) :: airmass      ! [     ] Air mass number (about 1/sin(elev))
     real (8) :: h2omm        ! [   mm] Zenith water vapor content
     type (opacity)   :: taus ! [neper] Zenith opacities at signal frequency
     type (opacity)   :: taui ! [neper] Zenith opacities at image  frequency
     type (dsb_value) :: temp ! [    K] Sky mean     temperatures (signal and image)
     type (dsb_value) :: temi ! [    K] Sky emission temperatures (signal and image)
  end type atm_prop
  !
end module chopper_definitions
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
